/* ========================================================================== */
/* Copyright (c) 2021 Michael Baeuerle
 *
 * All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, and/or sell copies of the Software, and to permit persons
 * to whom the Software is furnished to do so, provided that the above
 * copyright notice(s) and this permission notice appear in all copies of
 * the Software and that both the above copyright notice(s) and this
 * permission notice appear in supporting documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
 * OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY
 * SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Except as contained in this notice, the name of a copyright holder
 * shall not be used in advertising or otherwise to promote the sale, use
 * or other dealings in this Software without prior written authorization
 * of the copyright holder.
 */

/*! Tell headers that they should be POSIX compliant */
#define _POSIX_C_SOURCE  200112L


/* POSIX */
#include <sys/wait.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>

/* C99 */
#include <stdlib.h>
#include <string.h>

/* Local */
#include "canlock-hp.h"


/* ========================================================================== */
/*! \brief Write data into pipe to stdin of parser process
 *
 * \param[in,out] fd      Pointer to filedescriptor
 * \param[in]     data    Source buffer
 * \param[in]     len     Size of source buffer
 * \param[in,out] i       Index in source buffer
 *
 * \return
 * - 0 for operation in progress or finished (call again)
 * - -1 for error
 */
static int cl_hp_i_write_stdin(int *fd, const char *data, size_t len, size_t* i)
{
   ssize_t num;

   if (len > *i)
   {
      num = write(*fd, &data[*i], len - *i);
      if (-1 == num)
      {
         if (EINTR == errno || EAGAIN == errno)
            return 0;
         else
            return -1;
      }
      *i += (size_t) num;
   }
   else
   {
      /* No more data to write => Close write end of pipe at EOD */
      if (-1 != *fd)
      {
         if (0 != close(*fd))
            return -1;
         *fd = -1;
      }
   }

   return 0;
}


/* ========================================================================== */
/*! \brief Read data out of pipe from stdout of parser process
 *
 * \param[in]     fd      Filedescriptor
 * \param[out]    buffer  Target buffer
 * \param[in]     len     Size of target buffer
 * \param[in,out] i       Index in target buffer
 *
 * If there is not enough space in the target buffer, this is treated as error.
 * The target buffer should have the same size as the input data and the parser
 * output should never be larger than the input (because the parser extracts
 * only parts of the input data).
 *
 * \note
 * A NUL-termination was written into \e buffer if 1 (EOD) is returned.
 *
 * \return
 * - 0 for operation in progress (call again)
 * - 1 for EOD (no more data)
 * - -1 for error
 */
static int cl_hp_i_read_stdout(int fd, char *buffer, size_t len, size_t* i)
{
   ssize_t num = -1;

   if (len > *i)
   {
      num = read(fd, &buffer[*i], len - *i);
      if (-1 == num)
      {
         if (EINTR == errno || EAGAIN == errno)
            return 0;
         else
            return -1;
      }
      *i += (size_t) num;
   }
   else
      return -1;

   if (0 == num)
   {
      /* No more data to read => Write NUL-termination */
      if (len <= *i)
         return -1;
      buffer[*i] = 0;
      return 1;
   }

   return 0;
}


/* ========================================================================== */
/*! \brief Execute parser
 *
 * \param[in] pathname  Pathname of parser executable
 * \param[in] field     Option for header parser
 * \param[in] data      Input data to parse
 * \param[in] data_len  Length of input data
 *
 * \note
 * The parameter \e field should point to a string with the header field name
 * for the header parser \c canlock-mhp and should be \c NULL for the header
 * field parser \c canlock-hfp ).
 *
 * This wrapper function spawns a process for the parser specified by
 * \e pathname and feed \e data_len bytes, starting at the position pointed to
 * by \e data , into its standard input.
 *
 * \note \e data needs no NUL-termination at the end.
 *
 * The data from standard output is written to a new memory buffer allocated
 * with \c malloc() .
 *
 * The data from standard error is redirected to \c /dev/null .
 *
 * \return
 * Upon successful completion a pointer to the NUL-terminated result string is
 * returned. Otherwise, \c NULL is returned.
 * <br>
 * On success the caller is responsible to call free(3) for the pointer
 * returned as result.
 */
static char *cl_hp_parse(const char *pathname, const char *field,
                         const char *data, size_t data_len)
{
   pid_t pid, pid_rv;
   int stat_loc, input_fd[2], output_fd[2], error_fd, rv;
   char *buf;
   size_t ri = 0, wi = 0;  /* Read and write indexes */
   int failed = 0;

   /* Create pipes for stdin and stdout with nonblocking local endpoints */
   if (0 != pipe(input_fd))
      return NULL;
   if (0 != fcntl(input_fd[1], F_SETFL, O_NONBLOCK))
      return NULL;
   if (0 != pipe(output_fd))
      return NULL;
   if (0 != fcntl(output_fd[0], F_SETFL, O_NONBLOCK))
      return NULL;

   /* Spawn child process for parser */
   pid = fork();
   if (-1 == pid)
      return NULL;
   if (pid)
   {
      (void) close(input_fd[0]);
      (void) close(output_fd[1]);

      buf = (char *) malloc(data_len);
      if (NULL == buf)
         failed = 1;
      else
      {
         do
         {
            rv = cl_hp_i_write_stdin(&input_fd[1], data, data_len, &wi);
            if (0 == rv)
               rv = cl_hp_i_read_stdout(output_fd[0], buf, data_len, &ri);
         }
         while (!rv);
         if (0 > rv)
            failed = 1;
      }

      /* Explicitly terminate child process if an error occurred */
      if (failed)
         (void) kill(pid, SIGTERM);

      /* Reap exit status of child process */
      do
         pid_rv = waitpid(pid, &stat_loc, 0);
      while (-1 == pid_rv && EINTR == errno);
      if (pid != pid_rv)
         failed = 1;
      else if (0 != WEXITSTATUS(stat_loc))
         failed = 1;

      if (failed)
      {
         free((void *) buf);
         buf = NULL;
      }
      return buf;
   }
   else
   {
      /* -------------------------------------------------------------------- */
      /* Executed in new process */
      if (-1 == dup2(input_fd[0], STDIN_FILENO))
         exit(EXIT_FAILURE);
      (void) close(input_fd[0]);
      (void) close(input_fd[1]);

      if (-1 == dup2(output_fd[1], STDOUT_FILENO))
         exit(EXIT_FAILURE);
      (void) close(output_fd[0]);
      (void) close(output_fd[1]);

#if 1
      /* Redirect stderr to "/dev/null" */
      error_fd = open("/dev/null", O_WRONLY);
      if (-1 == error_fd)
         exit(EXIT_FAILURE);
      if (-1 == dup2(error_fd, STDERR_FILENO))
         exit(EXIT_FAILURE);
      (void) close(error_fd);
#endif

      /* Replace process image with the parser */
      do
         if (NULL != field)
            rv = execl(pathname, "canlock-mhp", "-f", field, (char *) NULL);
         else
            rv = execl(pathname, "canlock-hfp", (char *) NULL);
      while (-1 == rv && EINTR == errno);
      exit(EXIT_FAILURE);  /* Executed only if exec() has failed */
      /* -------------------------------------------------------------------- */
   }
}


/* ========================================================================== */
/*! \brief Execute header parser
 *
 * \param[in] data      Input data to parse
 * \param[in] data_len  Length of input data
 * \param[in] field     Header field to extract
 *
 * \note
 * \e data_len bytes are read, \e data needs no NUL-termination at the end.
 *
 * \return
 * Upon successful completion a pointer to the NUL-terminated result string is
 * returned. Otherwise, \c NULL is returned.
 * <br>
 * On success the caller is responsible to call free(3) for the pointer
 * returned as result.
 */
char *cl_hp_get_field(const char *data, size_t data_len,
                            const char *field)
{
   return cl_hp_parse("/usr/bin/canlock-mhp", field, data, data_len);
}


/* ========================================================================== */
/*! \brief Execute header field parser
 *
 * \param[in] data      Input data to parse
 * \param[in] data_len  Length of input data
 *
 * \note
 * \e data_len bytes are read, \e data needs no NUL-termination at the end.
 *
 * The input data must be an unfolded "Cancel-Lock" or "Cancel-Key" header field
 * without line break at the end.
 *
 * \return
 * Upon successful completion a pointer to the NUL-terminated result string is
 * returned. Otherwise, \c NULL is returned.
 * <br>
 * On success the caller is responsible to call free(3) for the pointer
 * returned as result.
 */
char *cl_hp_parse_field(const char *data, size_t data_len)
{
   return cl_hp_parse("/usr/bin/canlock-hfp", NULL, data, data_len);
}


/* EOF */
